#include "sdk_common.h"
#include "ble_srv_common.h"
#include "srm_service.h"
#include <string.h>
#include "nrf_gpio.h"
#include "boards.h"
#include "nrf_log.h"

static const uint8_t state_char_name[] = "State";
static const uint8_t cmd_char_name[] = "Command";
static const uint8_t data_char_name[] = "Data";

/**@brief Function for initializing the Custom Service.
 *
 * @param[out]  p_cus       Custom Service structure. This structure will have to be supplied by
 *                          the application. It will be initialized by this function, and will later
 *                          be used to identify this particular service instance.
 * @param[in]   p_srvc_opt_init  Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on successful initialization of service, otherwise an error code.
 */

//Adding the State Characteristic
static uint32_t state_char_add(ble_srvc_opt_t *p_cus, const ble_srvc_opt_init_t *p_srvc_opt_init, uint16_t uuid)
{
	uint32_t err_code  = NRF_SUCCESS;
	ble_gatts_char_md_t char_md;
	ble_gatts_attr_md_t cccd_md;
	ble_gatts_attr_t attr_char_value;
	ble_uuid_t ble_uuid;
	ble_gatts_attr_md_t attr_md;

	memset(&char_md, 0, sizeof(char_md));

	char_md.char_props.read = 1;
	char_md.char_props.write = 0;
	char_md.char_props.notify = 1;
	char_md.p_char_user_desc = state_char_name;
	char_md.char_user_desc_size = sizeof(state_char_name);
	char_md.char_user_desc_max_size = sizeof(state_char_name);
	char_md.p_char_pf = NULL;
	char_md.p_user_desc_md = NULL;
	char_md.p_cccd_md = NULL;
	char_md.p_sccd_md = NULL;

	memset(&attr_md, 0, sizeof(attr_md));

	attr_md.read_perm = p_srvc_opt_init->custom_value_char_attr_md.read_perm;
	attr_md.write_perm = p_srvc_opt_init->custom_value_char_attr_md.write_perm;
	attr_md.vloc = BLE_GATTS_VLOC_STACK;
	attr_md.rd_auth = 0;
	attr_md.wr_auth = 0;
	attr_md.vlen = 0;

	ble_uuid.type = p_cus->uuid_type;
	ble_uuid.uuid = BLE_SRM_STATE_CHAR_UUID;
	//ble_uuid.uuid = uuid;

	memset(&attr_char_value, 0, sizeof(attr_char_value));

	attr_char_value.p_uuid = &ble_uuid;
	attr_char_value.p_attr_md = &attr_md;
	attr_char_value.init_len = sizeof(uint8_t);
	attr_char_value.init_offs = 0;
	attr_char_value.max_len = sizeof(uint8_t);
	uint8_t value[1] = {0x00};
	attr_char_value.p_value = value;

	memset(&cccd_md, 0, sizeof(cccd_md));

	//  Read  operation on Cccd should be possible without authentication.
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);

	cccd_md.vloc = BLE_GATTS_VLOC_STACK;

	char_md.p_cccd_md = &cccd_md;

	err_code = sd_ble_gatts_characteristic_add(p_cus->service_handle, &char_md,
											   &attr_char_value,
											   &p_cus->state_char_handle);
	
	if (err_code != NRF_SUCCESS){return err_code;}
	return err_code;
}

//Adding the Command Characteristic
static uint32_t cmd_char_add(ble_srvc_opt_t *p_cus, const ble_srvc_opt_init_t *p_srvc_opt_init, uint16_t uuid)
{
	uint32_t err_code  = NRF_SUCCESS;
	ble_gatts_char_md_t char_md;
	ble_gatts_attr_md_t cccd_md;
	ble_gatts_attr_t attr_char_value;
	ble_uuid_t ble_uuid;
	ble_gatts_attr_md_t attr_md;

	memset(&char_md, 0, sizeof(char_md));

	char_md.char_props.read = 1;
	char_md.char_props.write = 1;
	char_md.char_props.notify = 0;
	char_md.p_char_user_desc = cmd_char_name;
	char_md.char_user_desc_size = sizeof(cmd_char_name);
	char_md.char_user_desc_max_size = sizeof(cmd_char_name);
	char_md.p_char_pf = NULL;
	char_md.p_user_desc_md = NULL;
	char_md.p_cccd_md = NULL;
	char_md.p_sccd_md = NULL;

	memset(&attr_md, 0, sizeof(attr_md));

	attr_md.read_perm = p_srvc_opt_init->custom_value_char_attr_md.read_perm;
	attr_md.write_perm = p_srvc_opt_init->custom_value_char_attr_md.write_perm;
	attr_md.vloc = BLE_GATTS_VLOC_STACK;
	attr_md.rd_auth = 0;
	attr_md.wr_auth = 0;
	attr_md.vlen = 0;

	ble_uuid.type = p_cus->uuid_type;
	ble_uuid.uuid = BLE_SRM_CMD_CHAR_UUID;

	memset(&attr_char_value, 0, sizeof(attr_char_value));

	attr_char_value.p_uuid = &ble_uuid;
	attr_char_value.p_attr_md = &attr_md;
	attr_char_value.init_len = sizeof(uint8_t);
	attr_char_value.init_offs = 0;
	attr_char_value.max_len = sizeof(uint8_t);
	uint8_t value[1] = {0x00};
	attr_char_value.p_value = value;

	memset(&cccd_md, 0, sizeof(cccd_md));

	//  Read  operation on Cccd should be possible without authentication.
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);

	cccd_md.vloc = BLE_GATTS_VLOC_STACK;

	char_md.p_cccd_md = &cccd_md;

	err_code = sd_ble_gatts_characteristic_add(p_cus->service_handle, &char_md,
											   &attr_char_value,
											   &p_cus->cmd_char_handle);
	if (err_code != NRF_SUCCESS){return err_code;}
	return err_code;
}

//Adding the Data Characteristic
static uint32_t data_char_add(ble_srvc_opt_t *p_cus, const ble_srvc_opt_init_t *p_srvc_opt_init, uint16_t uuid)
{
	uint32_t err_code;
	ble_gatts_char_md_t char_md;
	ble_gatts_attr_md_t cccd_md;
	ble_gatts_attr_t attr_char_value;
	ble_uuid_t ble_uuid;
	ble_gatts_attr_md_t attr_md;

	memset(&char_md, 0, sizeof(char_md));

	char_md.char_props.read = 1;
	char_md.char_props.write = 0;
	char_md.char_props.notify = 1;
	char_md.p_char_user_desc = data_char_name;
	char_md.char_user_desc_size = sizeof(data_char_name);
	char_md.char_user_desc_max_size = sizeof(data_char_name);
	char_md.p_char_pf = NULL;
	char_md.p_user_desc_md = NULL;
	char_md.p_cccd_md = NULL;
	char_md.p_sccd_md = NULL;

	memset(&attr_md, 0, sizeof(attr_md));

	attr_md.read_perm = p_srvc_opt_init->custom_value_char_attr_md.read_perm;
	attr_md.write_perm = p_srvc_opt_init->custom_value_char_attr_md.write_perm;
	attr_md.vloc = BLE_GATTS_VLOC_STACK;
	attr_md.rd_auth = 0;
	attr_md.wr_auth = 0;
	attr_md.vlen = 0;

	ble_uuid.type = p_cus->uuid_type;
	ble_uuid.uuid = BLE_SRM_CMD_CHAR_UUID;

	memset(&attr_char_value, 0, sizeof(attr_char_value));

	attr_char_value.p_uuid = &ble_uuid;
	attr_char_value.p_attr_md = &attr_md;
	attr_char_value.init_len = sizeof(data_values_t);
	//attr_char_value.init_offs = 0;
	attr_char_value.max_len = sizeof(data_values_t);
	uint8_t value[22] = {	0x1A, 0x1A, 0x1B, 0x1C, //unit32_t serial;
							0x2A, 0x2B, //int16_t val_a;
    						0x3A, 0x3B, //int16_t val_b;
    						0x4A, 0x4B, //int16_t val_c;
							0x5A, 0x5B, //int16_t val_d;
    						0x6A, 0x6B, //int16_t val_e;
    						0x7A, 0x7B, //int16_t val_f;
							0x8A, 0x8B, //int16_t val_g;
    						0x9A, 0x9A, //int16_t val_h;
    						0xAA, 0xAB, //int16_t val_i;
						};
	attr_char_value.p_value = value;
	memset(&cccd_md, 0, sizeof(cccd_md));

	//  Read  operation on Cccd should be possible without authentication.
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);

	cccd_md.vloc = BLE_GATTS_VLOC_STACK;

	char_md.p_cccd_md = &cccd_md;

	err_code = sd_ble_gatts_characteristic_add(p_cus->service_handle, &char_md,
											   &attr_char_value,
											   &p_cus->data_char_handle);
	if (err_code != NRF_SUCCESS)
	{
		return err_code;
	}

	return NRF_SUCCESS;
}

//Initialise the Service and add the Characteristics
uint32_t ble_srvc_opt_init(ble_srvc_opt_t *p_cus, const ble_srvc_opt_init_t *p_srvc_opt_init)
{

	if (p_cus == NULL || p_srvc_opt_init == NULL)
	{
		return NRF_ERROR_NULL;
	}

	uint32_t err_code  = NRF_SUCCESS;
	ble_uuid_t ble_uuid;
	uint16_t cus_char_uuid;

	// Initialize service structure
	p_cus->evt_handler = p_srvc_opt_init->evt_handler;
	p_cus->conn_handle = BLE_CONN_HANDLE_INVALID;
	p_cus->is_notification_enabled = false;

	// Add Custom Service UUID
	ble_uuid128_t base_uuid = {BLE_SRM_SERVICE_BASE_UUID};
	err_code = sd_ble_uuid_vs_add(&base_uuid, &p_cus->uuid_type);
	VERIFY_SUCCESS(err_code);

	ble_uuid.type = p_cus->uuid_type;
	ble_uuid.uuid = BLE_SRM_SERVICE_UUID;

	// Add the Custom Service
	err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_cus->service_handle);
	if (err_code != NRF_SUCCESS){return err_code;}

	// Add the State Characteristic
	err_code = state_char_add(p_cus, p_srvc_opt_init, cus_char_uuid);
	if (err_code != NRF_SUCCESS){return err_code;}

    // Add the Command Characteristic
	err_code = cmd_char_add(p_cus, p_srvc_opt_init, cus_char_uuid);
	if (err_code != NRF_SUCCESS){return err_code;}

    // Add the Data Characteristic
	err_code = data_char_add(p_cus, p_srvc_opt_init, cus_char_uuid);
	if (err_code != NRF_SUCCESS){return err_code;}

	return err_code;
}

/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_srvc_opt_t *p_cus, ble_evt_t const *p_ble_evt)
{
	p_cus->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

	ble_srvc_opt_evt_t evt;

	evt.evt_type = BLE_SRVC_EVT_CONNECTED;

	p_cus->evt_handler(p_cus, &evt);

}

/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_srvc_opt_t *p_cus, ble_evt_t const *p_ble_evt)
{
	UNUSED_PARAMETER(p_ble_evt);
	p_cus->conn_handle = BLE_CONN_HANDLE_INVALID;
	//nrf_gpio_pin_toggle(LED_4);
}

static void on_write(ble_srvc_opt_t *p_cus, ble_evt_t const *p_ble_evt)
{
	//ble_gatts_evt_write_t * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;
	ble_gatts_evt_write_t const *p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

	// State Characteristic Written to.
	if (p_evt_write->handle == p_cus->state_char_handle.value_handle)
	{
		if (p_cus->evt_handler != NULL)
		{
			//Set and Call the application event handler
			ble_srvc_opt_evt_t evt;
			evt.evt_type = BLE_SRVC_EVT_STATE_CHAR_WRITE;
			p_cus->evt_handler(p_cus, &evt);

		}
	}

    // Command Characteristic Written to.
	if (p_evt_write->handle == p_cus->cmd_char_handle.value_handle)
	{
		if (p_cus->evt_handler != NULL)
		{
			//Set and Call the application event handler
			ble_srvc_opt_evt_t evt;
			evt.evt_type = BLE_SRVC_EVT_CMD_CHAR_WRITE;
			p_cus->evt_handler(p_cus, &evt);

		}
	}

	// Check if the Custom value CCCD is written to and that the value is the appropriate length, i.e 2 bytes.
	if ((p_evt_write->handle == p_cus->state_char_handle.cccd_handle) && (p_evt_write->len == 2))
	{
		// CCCD written, call application event handler
		if (p_cus->evt_handler != NULL)
		{
			ble_srvc_opt_evt_t evt;

			if (ble_srv_is_notification_enabled(p_evt_write->data))
			{
				evt.evt_type = BLE_SRVC_EVT_STATE_NOTIFICATION_ENABLED;
				p_cus->is_notification_enabled = true;
				//NRF_LOG_INFO("ble_srvc_opt- on_write - BLE_SRVC_OPT_EVT_NOTIFICATION_ENABLED");
			}
			else
			{
				evt.evt_type = BLE_SRVC_EVT_STATE_NOTIFICATION_DISABLED;
				p_cus->is_notification_enabled = false;
				//NRF_LOG_INFO("ble_srvc_opt - on_write - BLE_SRVC_OPT_EVT_NOTIFICATION_DISABLED");
			}
			// Call the application event handler.
			p_cus->evt_handler(p_cus, &evt);
		}
	}
}

void ble_srvc_opt_on_ble_evt(ble_evt_t const *p_ble_evt, void *p_context)
{
	ble_srvc_opt_t *p_cus = (ble_srvc_opt_t *)p_context;

	if (p_cus == NULL || p_ble_evt == NULL)
	{
		return;
	}

	switch (p_ble_evt->header.evt_id)
	{
	case BLE_GAP_EVT_CONNECTED:
		NRF_LOG_INFO("ble_srvc_opt - BLE_GAP_EVT_CONNECTED");
		on_connect(p_cus, p_ble_evt);
		break;

	case BLE_GAP_EVT_DISCONNECTED:
		NRF_LOG_INFO("ble_srvc_opt - BLE_GAP_EVT_DISCONNECTED");
		on_disconnect(p_cus, p_ble_evt);
		break;
	case BLE_GATTS_EVT_WRITE:
		on_write(p_cus, p_ble_evt);
		break;
	default:
		// No implementation needed.
		break;
	}
}

//Function for setting the State Characteristic
uint32_t ble_srvc_set_state(ble_srvc_opt_t *p_service, uint8_t state_value) {

	if (p_service == NULL) {return NRF_ERROR_NULL;}
	uint32_t err_code = NRF_SUCCESS;

	//Setup the value to send
	ble_gatts_value_t	t_value;
    t_value.p_value = &state_value;
	t_value.len      = 1;
	t_value.offset   = 0;

	//setup the notification event
	ble_gatts_hvx_params_t hvx_params;
    memset(&hvx_params, 0, sizeof(hvx_params));
    hvx_params.handle = p_service->state_char_handle.value_handle;
	hvx_params.type = BLE_GATT_HVX_NOTIFICATION;
    hvx_params.offset = t_value.offset;
	hvx_params.p_len = &t_value.len;

	//Set value
	err_code = sd_ble_gatts_value_set(p_service->conn_handle, p_service->state_char_handle.value_handle, &t_value);
	if (err_code != NRF_SUCCESS) {return err_code;}

	//Set notification
    if ( p_service->conn_handle != BLE_CONN_HANDLE_INVALID )
	{
		hvx_params.p_data = t_value.p_value;
		err_code = sd_ble_gatts_hvx(p_service->conn_handle, &hvx_params);
		if (err_code != NRF_SUCCESS) {return err_code;}
	}
	return err_code;

}

//Function for setting the Option Characteristic
uint32_t ble_srvc_opt_custom_value_update(ble_srvc_opt_t *p_cus, uint8_t custom_value)
{
	NRF_LOG_INFO("In ble_srvc_opt_custom_value_update. \r\n");
	if (p_cus == NULL) {return NRF_ERROR_NULL;}

	uint32_t err_code = NRF_SUCCESS;
	ble_gatts_value_t gatts_value;

	// Initialize value struct.
	memset(&gatts_value, 0, sizeof(gatts_value));

	gatts_value.len = sizeof(uint8_t);
	gatts_value.offset = 0;
	gatts_value.p_value = &custom_value;

	// Update database.
	err_code = sd_ble_gatts_value_set(p_cus->conn_handle,
									  p_cus->state_char_handle.value_handle,
									  &gatts_value);
	if (err_code != NRF_SUCCESS) {return err_code;}

	// Send value if connected and notifying.
	if ((p_cus->conn_handle != BLE_CONN_HANDLE_INVALID))
	{
		ble_gatts_hvx_params_t hvx_params;

		memset(&hvx_params, 0, sizeof(hvx_params));

		hvx_params.handle = p_cus->state_char_handle.value_handle;
		hvx_params.type = BLE_GATT_HVX_NOTIFICATION;
		hvx_params.offset = gatts_value.offset;
		hvx_params.p_len = &gatts_value.len;
		hvx_params.p_data = gatts_value.p_value;

		err_code = sd_ble_gatts_hvx(p_cus->conn_handle, &hvx_params);
		if (err_code != NRF_SUCCESS) {return err_code;}
	}
	return err_code;
}

//Function for setting the MPU Characteristic
uint32_t ble_mpu_update(ble_srvc_opt_t *p_cus, data_values_t * send_value)
{
    
	if (&p_cus == NULL) {return NRF_ERROR_NULL;}
	uint32_t err_code = NRF_SUCCESS;

	//Setup the value to send
	uint8_t value[22] = {

		(send_value->serial & 0xFF000000) >> 24,
		(send_value->serial & 0x00FF0000) >> 16,
		(send_value->serial & 0x0000FF00) >> 8,
		(send_value->serial & 0x000000FF),

		(send_value->val_a & 0xFF00) >> 8,
		(send_value->val_a & 0x00FF),
		
		(send_value->val_b & 0xFF00) >> 8,
		(send_value->val_b & 0x00FF),
		
		(send_value->val_c & 0xFF00) >> 8,
		(send_value->val_c & 0x00FF),
		
		(send_value->val_d & 0xFF00) >> 8,
		(send_value->val_d & 0x00FF),
		
		(send_value->val_e & 0xFF00) >> 8,
		(send_value->val_e & 0x00FF),
		
		(send_value->val_f & 0xFF00) >> 8,
		(send_value->val_f & 0x00FF),
		
		(send_value->val_g & 0xFF00) >> 8,
		(send_value->val_g & 0x00FF),
		
		(send_value->val_h & 0xFF00) >> 8,
		(send_value->val_h & 0x00FF),
		
		(send_value->val_i & 0xFF00) >> 8,
		(send_value->val_i & 0x00FF),
		
	};
	ble_gatts_value_t	t_value;
    t_value.p_value = value;
	t_value.len      = sizeof(data_values_t);
	t_value.offset   = 0;

	//setup the notification event
	ble_gatts_hvx_params_t hvx_params;
    memset(&hvx_params, 0, sizeof(hvx_params));
    hvx_params.handle = p_cus->data_char_handle.value_handle;
	hvx_params.type = BLE_GATT_HVX_NOTIFICATION;
    hvx_params.offset = t_value.offset;
	hvx_params.p_len = &t_value.len;

	//Set value
	err_code = sd_ble_gatts_value_set(p_cus->conn_handle, p_cus->data_char_handle.value_handle, &t_value);
    if (err_code != NRF_SUCCESS) {return err_code;}

	//Set notification
    if ((p_cus->conn_handle != BLE_CONN_HANDLE_INVALID))
	{
		hvx_params.p_data = t_value.p_value;
		err_code = sd_ble_gatts_hvx(p_cus->conn_handle, &hvx_params);
		if (err_code != NRF_SUCCESS) {return err_code;}
	}
	return err_code;

}